/*----------------------------------------------------------------------------
 *
 *  main.c
 *  ------
 *
 *  Purpose: Provide the startup processing for VisualAge Smalltalk Enterprise
 *
 *  Commandline switches:
 *   -nosplash  Do not display the splash screen
 *   -singleinstance  Disallow multiple copies of the executable running
 *                    at the same time
 *   -no_break  Do not display the User Break button
 *
 *  Licensed Materials - Property of IBM
 *
 *  (C) Copyright International Business Machines Corp., 1999
 *  All rights reserved
 *
 *  US Government Users Restricted Rights -
 *  Use, duplication, or disclosure restricted
 *  by GSA ADP Schedule Contract with IBM Corp.
 *
 *  Modifications:
 *
 *    5.5.0.0 11Nov99(jok) Update to 5.1.0 w/2000 copyright
 *    5.5.0.1 19Jan00(jok) Load bitmap from file
 *    5.5.0.2 02Mar00(jok) ES40MD30 rebuild
 *
 *----------------------------------------------------------------------------*/

#define PRODUCT_NAME    "VisualAge Smalltalk Enterprise"
#define PRODUCT_VERSION "Version 5.5"
#define PRODUCT_STATUS  ""
#define PRODUCT_EDITION "03/23/2000"
#define PRODUCT_COPYR   "(C) Copyright International Business Machines Corporation  1994, 2000.\nAll Rights Reserved."

#include <windows.h>
#include <stdio.h>
#include <string.h>
#include "esrc.h"
#include "esuser.h"

/*--------------------------------------------------------------------------*/
/*      DEBUG support                                                       */
/*--------------------------------------------------------------------------*/

#ifdef DEBUG01
#define dbg_printf( x ) { printf x ; fflush( stdout ); }
#else
#define dbg_printf( x )
#endif

#define QUOTE_CHAR '"'

   U_32 WindowsVersion;

   static HINSTANCE	hInstanceCur;
   static WNDPROC StaticProc;
   static HBITMAP hwndSplashBitmap = (HBITMAP) NULL;
   static BITMAP bitmap;
   static char *ES_cmdLineParams[128];

   /* function prototypes */
   BOOLEAN EsFindImageFileName( int argc, char *argv[], char *envp[], EsGlobalInfo * info );
   BOOLEAN EsParseCommandLineOptions( int argc, char *argv[], char *envp[], EsGlobalInfo * info );
   void EsSetGlobalHInstance( ESGlobalInfo *ginfo, HINSTANCE hInst );
   extern void EsSplashShutDown( EsGlobalInfo* );
   extern EsPrimitiveTable WindowProcPrimitiveTable;

   EsDefinePrimitiveTable(EsExePrimitiveTable)
      EsSubTable(WindowProcPrimitiveTable)
   EsEndPrimitiveTable



/**************************************************************************
*  Takes the command line (1 string) and breaks it into an argv[] style
*  list.
*
*  Understands Win95 LFNs and strips quotes off the exe name and the
*  parameters.
***************************************************************************/

parseWords(char *buffer, char *words[]) {
   int count = 0;

   while (*buffer) {
      if (QUOTE_CHAR == *buffer) {	/* we have a quoted name. */
         words[count++] = ++buffer;	/* move past the quote */
         while (*buffer && QUOTE_CHAR != *buffer)
            buffer++;
         if (*buffer)
            *buffer++ = '\0';	/* slam the close quote and advance */
         }
      else {
         words[count++] = buffer;
         while (*buffer && *buffer != ' ' && *buffer != '\t' && *buffer != '\r')
            buffer++;
         }
      if (*buffer)
         *buffer++ = '\0';
      while (*buffer == ' ' || *buffer == '\t' || *buffer == '\r')
         buffer++;
      }
   return count;
   }



/**************************************************************************
* 
*  function:  LoadBitmapFile()
*   Load a .bmp file with the same name as the .exe file
* 
*  Input: 
*   hDC - hDC to make the bitmap compatible with. 
* 
*  Returns:
*   handle to the bitmap if successful; otherwise NULL. 
* 
***************************************************************************/ 

HBITMAP LoadBitmapFile( HDC hDC ) { 
   HBITMAP	hBitmap = NULL;		/* Setup for error exit */ 
   BITMAPFILEHEADER	fileHeader; 
   PBITMAPINFOHEADER	pInfoHeader; 
   PBYTE	pBuffer; 
   HANDLE	hFile; 
   DWORD	bytesToRead;
   DWORD	bytesActuallyRead;
   BOOL		rc;
   char		bitmapPath[MAX_PATH]; /* path to <splash>.bmp */ 
   int		len;

   dbg_printf( ("Entered LoadBitmapFile\n") );
   len = strlen( ES_cmdLineParams[0] );
   if ((len < 5) || (len > MAX_PATH) || strcmpi( ES_cmdLineParams[0] + len - 4, ".exe" ))
      return NULL;	/* Executable name is not properly formed */
   strcpy( bitmapPath, ES_cmdLineParams[0] );
   strcpy( bitmapPath + len - 3, "bmp" );
   dbg_printf( ("Showing splash from '%s'\n", bitmapPath) );
	
   hFile = CreateFile( (LPCTSTR) bitmapPath,
                       GENERIC_READ,
                       FILE_SHARE_READ,
                       NULL,			
                       OPEN_EXISTING,
                       FILE_ATTRIBUTE_READONLY | FILE_FLAG_SEQUENTIAL_SCAN,
                       NULL);			
   if (hFile != INVALID_HANDLE_VALUE ) {
      bytesToRead = sizeof( BITMAPFILEHEADER );
      rc = ReadFile( hFile,
                     (LPVOID) &fileHeader,
                     bytesToRead,
                     &bytesActuallyRead,
                     (LPOVERLAPPED) NULL );
      if ((rc != 0) & 
          (bytesToRead == bytesActuallyRead) &
          (fileHeader.bfType == 'MB') & 
          (fileHeader.bfSize == GetFileSize( hFile, NULL))) {
         bytesToRead = fileHeader.bfOffBits - sizeof( BITMAPFILEHEADER );
         pInfoHeader = (PBITMAPINFOHEADER) LocalAlloc( LPTR, bytesToRead );
         if (pInfoHeader != NULL) {
            rc = ReadFile( hFile,
                           (LPVOID) pInfoHeader,
                           bytesToRead,
                           &bytesActuallyRead,
                           (LPOVERLAPPED) NULL );
            if ((rc != 0) &
                (bytesToRead == bytesActuallyRead)) {
               bytesToRead = fileHeader.bfSize - fileHeader.bfOffBits;
               pBuffer = (PBYTE) LocalAlloc( LPTR, bytesToRead );
               if (pBuffer != NULL) {
                  dbg_printf( ("\tBitmap memory allocated\n") );
                  rc = ReadFile( hFile,
                                 (LPVOID) pBuffer,
                                 bytesToRead,
                                 &bytesActuallyRead,
                                 (LPOVERLAPPED) NULL );
                  if ((rc !=0) &
                      (bytesToRead == bytesActuallyRead))
                     dbg_printf( ("\tBitmap file read is OK\n") );
                     hBitmap = CreateDIBitmap( hDC, 
                                               (PBITMAPINFOHEADER) pInfoHeader, 
                                               CBM_INIT, 
                                               pBuffer, 
                                               (PBITMAPINFO) pInfoHeader, 
                                               DIB_RGB_COLORS); 
                  LocalFree( LocalHandle( (LPSTR) pBuffer ) );
                  } 
               }
            LocalFree( LocalHandle( (LPSTR) pInfoHeader ) );
            }
         } 
      CloseHandle( hFile );
      }
   return hBitmap; 
   } 



LRESULT CALLBACK DrawBitmapProc( HWND hwndBitmap, UINT msg, WPARAM wParam, LPARAM lParam ) {				
   HBITMAP hOldBitmap;
   HDC hMemDC, hDC;
   PAINTSTRUCT ps;

   dbg_printf( ("Entered DrawBitmapProc\n") );
   switch(msg) {
      case WM_PAINT: {
         dbg_printf( ("\tGot WM_PAINT\n") );
         if (hwndSplashBitmap != NULL) {
            hDC = BeginPaint( hwndBitmap, &ps );
            hMemDC = CreateCompatibleDC( hDC );
            hOldBitmap = SelectObject( hMemDC, hwndSplashBitmap );
            BitBlt( hDC,
                    0,
                    0,
                    bitmap.bmWidth,
                    bitmap.bmHeight,
                    hMemDC,
                    0,
                    0,
                    SRCCOPY );
            SelectObject( hMemDC, hOldBitmap );
            DeleteDC( hMemDC );
            EndPaint( hwndBitmap, &ps );
            return (LRESULT) 0;
            }
         }
      }
   return (LRESULT) CallWindowProc( StaticProc, hwndBitmap, msg, wParam, lParam );
   }



BOOL CALLBACK DlgProc( HWND hwndDlg, UINT uMsg, WPARAM wParam, LPARAM lParam ) {
   int x, y;
   int size;

   dbg_printf( ("Entered DlgProc\n") );
   switch(uMsg) {
      case WM_INITDIALOG: {
         dbg_printf( ("\tGot Wm_INITDLG\n") );
         hwndSplashBitmap = LoadBitmapFile( GetDC( hwndDlg ) );
         if (hwndSplashBitmap != NULL)
            size = GetObject( hwndSplashBitmap, sizeof( BITMAP ), (LPSTR) &bitmap );
            if (size != 0) {
               x = max( (GetSystemMetrics( SM_CXSCREEN ) - bitmap.bmWidth) / 2, 0 );
               y = max( (GetSystemMetrics( SM_CYSCREEN ) - bitmap.bmHeight) / 2, 0 );
               MoveWindow( hwndDlg, x, y, bitmap.bmWidth, bitmap.bmHeight, TRUE );
               return TRUE;
               }
          break;

      case WM_CLOSE:
         DestroyWindow( hwndDlg );
         return TRUE;
         break;
			
      case WM_DESTROY:
         PostQuitMessage( 0 );
         return TRUE;
         break;
         }
      }
   return FALSE;
   }



static void splashDialogThread( EsGlobalInfo *globalInfo ) {
   HWND hwndStatic;
   HWND hwndDlgBox;
   MSG msg;

   dbg_printf( ("Entered splashDialogThread\n") );
   SetThreadPriority( GetCurrentThread(), THREAD_PRIORITY_ABOVE_NORMAL );
   hwndDlgBox = CreateDialog( hInstanceCur, "IDD_SPLASHDIALOG", NULL, DlgProc );
   if (hwndDlgBox != NULL) {    /* Got the dialog box handle */
      dbg_printf( ("\thwndDlgBox: %x\n", hwndDlgBox) );
      globalInfo->splashDialogHandle = hwndDlgBox;    /* Set VM handle to the dialog box */
      globalInfo->splashStatusLineHandle = NULL;    /* GetDlgItem (hwndDlgBox , IDS_STATUS); */
      globalInfo->messageLevel = ESVM_MSG_LEVEL_NONE;
      hwndStatic = GetDlgItem ( hwndDlgBox, IDBM_SPLASHBITMAP );
      if (hwndStatic != NULL) {
         dbg_printf( ("\thwndStatic: %x\n", hwndStatic) );
         StaticProc = (WNDPROC) GetWindowLong( hwndStatic, GWL_WNDPROC );
         SetWindowLong( hwndStatic, GWL_WNDPROC, (LONG) DrawBitmapProc );
         ShowWindow( hwndDlgBox, SW_SHOW );
         UpdateWindow( hwndDlgBox );
         /* NOTE: The splash window gets destroyed in DlgProc due to an
            EsSplashShutDown call from either the Smalltalk image
            (System>>#splashShutDown) or the main thread (in the case
            of an error starting Smalltalk).  This will cause the
            following while loop to terminate. */ 
         while (GetMessage( &msg, NULL, 0L, 0L )) {
            DispatchMessage( &msg );
            }
         }
      if (hwndSplashBitmap != NULL) {
         DeleteObject( hwndSplashBitmap );
         hwndSplashBitmap = NULL;
         }
      globalInfo->splashDialogHandle = NULL;
      globalInfo->splashStatusLineHandle = NULL;
      }
   ExitThread( 0 );
   }



U_32 GetPlatform(void) {
   OSVERSIONINFO osVerInfo;

   osVerInfo.dwOSVersionInfoSize=sizeof( OSVERSIONINFO );
   if (!GetVersionEx( &osVerInfo ))
      return VER_PLATFORM_WIN32s;
   return osVerInfo.dwPlatformId;
}



EsMain(int argc, char ** argv, char ** envp)
{
   I_32 rc, i;
   EsGlobalInfo *globalInfo;
   DWORD dwTid;
   HANDLE hThread;
   BOOLEAN splashEnabled = TRUE;
   BOOLEAN mutexEnabled = FALSE;
   void *startupMutex;
   char mutexNameBuffer[1024], *exeName, *exeExtension, *ptr;

   /* Walk the command line looking VA specific command line options */
   dbg_printf( ("Entering main, %d commandline arguments\n", argc) );
   for(i=0; i < argc; i++) {
      dbg_printf( ("\t%s\n", argv[i]) );
      if (strcmp(argv[i],"-nosplash") == 0)
         splashEnabled = FALSE;
      if (strcmp(argv[i],"-singleinstance") == 0)
         mutexEnabled = TRUE;
	}
   dbg_printf( ("splashEnabled: %d\n", splashEnabled) );
   dbg_printf( ("mutexEnabled: %d\n", mutexEnabled) );
	
   /* single start mutex acquisition */
   if (mutexEnabled) {
      /* strip the path name and extension from the file */
      exeName = strrchr(argv[0], '\\');
      if(exeName)
         exeName++;
      else
         exeName = argv[0];

      strcpy(mutexNameBuffer, exeName);
      for (ptr = mutexNameBuffer; *ptr; ptr++)
         *ptr = toupper(*ptr);
      exeExtension = strrchr(mutexNameBuffer, '.');
      if (exeExtension && 0 == strcmp( exeExtension, ".EXE" ))
         *exeExtension = '\0';
      strcat(mutexNameBuffer, "_SINGLE_START_MUTEX");

      SetLastError(0);	/* reset the Win32 error status */
      startupMutex = CreateMutex(NULL, FALSE, mutexNameBuffer);
      if (ERROR_ALREADY_EXISTS == GetLastError( )) {	/* mutex already exists */
         CloseHandle(startupMutex);
         return 0;
         }
      }

   WindowsVersion = GetPlatform();
   globalInfo = EsInitializeImage();
   if (globalInfo == NULL) {
      EsReportError( EsPrimErrNotEnoughMemory, globalInfo );
      if (mutexEnabled)
         CloseHandle( startupMutex );
      return EsPrimErrNotEnoughMemory;
      }
   globalInfo->imagePrimitives = (U_32) EsExePrimitiveTable;

   /* bring up the splash dialog, need globalInfo in hand first */
   if (splashEnabled) {
      hThread = CreateThread( NULL,
                              0,	
                              (LPTHREAD_START_ROUTINE) splashDialogThread,
                              globalInfo,
                              0,
                              &dwTid );
      if (hThread != NULL)
         CloseHandle( hThread );
      }

   EsSetGlobalHInstance( globalInfo, hInstanceCur );
			
   if ((rc = EsInitializeTargetInterface( argc, argv, envp, globalInfo )) != EsPrimErrNoError) {
      EsSplashShutDown( globalInfo );
      EsReportError( rc, globalInfo );
      EsShutdownImage( globalInfo );
      if (mutexEnabled)
         CloseHandle( startupMutex );
      return rc;
      }

   EsPrintf( "\n%s, %s %s - %s\nVM Timestamp: %s\n%s\n",
             PRODUCT_NAME,PRODUCT_VERSION,PRODUCT_STATUS,PRODUCT_EDITION,
             EsVMVersionString(),
             PRODUCT_COPYR,
             0,0,0,0);

   if (!EsParseCommandLineOptions( argc, argv, envp, globalInfo )) {
      EsSplashShutDown(globalInfo);
      EsReportError(EsPrimErrImageFileOpenError, globalInfo);
      EsShutdownImage(globalInfo);
      if (mutexEnabled)
         CloseHandle( startupMutex );
      return EsPrimErrImageFileOpenError;
      }

   rc = EsLoadFileImage( argc, argv, envp, globalInfo );
   if (rc == EsPrimErrNoError)
      rc = EsExecuteImage( globalInfo );
   if (rc != EsPrimErrNoError) {
      if (rc < 9000 || (rc > 9100))
         EsReportError( rc, globalInfo );
      }

   EsShutDownTargetInterface (globalInfo);
   EsShutdownImage(globalInfo);
   if (mutexEnabled)
      CloseHandle( startupMutex );
   return rc;
   }



WINAPI WinMain(	HINSTANCE hInstCurrent, HINSTANCE hInstPrevious, LPSTR lpszCmdLine, int nCmdShow) {
   int nArgs, pathLength, rc;
   char pgmName[MAX_PATH];

   void EsSetHInstance(HINSTANCE);
	
   hInstanceCur = hInstCurrent;
   EsSetHInstance(hInstCurrent);
   
   nArgs = parseWords(GetCommandLine(), &ES_cmdLineParams[0]);
   /* On Windows NT, argv[0] may only contain partial info, so
      we recompute it on all platforms */
   pathLength = GetModuleFileName( hInstCurrent,
                                    (LPSTR)&pgmName[0],
                                    MAX_PATH );
   if (pathLength != 0)
      ES_cmdLineParams[0] = &pgmName[0];
  
   rc = EsMain(nArgs, ES_cmdLineParams, NULL);
   return rc;
   }
